(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 25-Apr-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: BasicHLSL.cpp
//
// Desc: This sample shows a simple example of the Microsoft Direct3D's High-Level
//       Shader Language (HLSL) using the Effect interface.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit MultiAnimationUnit;

interface

uses
  Windows, Messages,
  Direct3D9, D3DX9, dxerr9, DirectSound,
  D3DApp, D3DFont, D3DUtil, DXUtil, DSUtil,
  Resource, MultiAnimationLib, Tiny;

// C++Builder compile hack
{$NOINCLUDE Resource}
{$HPPEMIT '#include "Resource.hpp"'}

{.$DEFINE DEBUG_VS}   // Uncomment this line to debug vertex shaders
{.$DEFINE DEBUG_PS}   // Uncomment this line to debug pixel shaders

const
  TXFILE_FLOOR = 'Floor.jpg';
  FLOOR_TILECOUNT = 2;

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  protected
    m_pMeshFloor:         ID3DXMesh;                  // floor geometry
    m_mxFloor:            TD3DXMatrix;                // floor world xform
    m_MatFloor:           TD3DMaterial9;              // floor material
    m_pTxFloor:           IDirect3DTexture9;          // floor texture

    m_pFont:              ID3DXFont;                  // framework font
    m_Camera:             CFirstPersonCamera;         // Camera for interactive character nagivation
    m_fAspectRatio:       Single;                     // Aspect ratio used by the FPS camera
    m_DSound:             CSoundManager;              // DSound class

    m_MultiAnim:          CMultiAnim;                 // the MultiAnim class for holding Tiny's mesh and frame hierarchy
    m_v_pCharacters:      TTinyList;                  // array of character objects; each can be associated with any of the CMultiAnims

    m_dwFollow:           DWORD;                      // which character the camera should follow; 0xffffffff for static camera
    m_bDrawReflection:    Boolean;                    // reflection-drawing option
    m_bDisplayText:       Boolean;                    // whether to display text
    m_bDisplayHelp:       Boolean;                    // whether to display help text

    m_bPlaySounds:        bool;                       // whether to play sounds

    m_fLastAnimTime:      Single;                     // Time for the animations

  protected
    function OneTimeSceneInit: HRESULT; override;
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function FinalCleanup: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

    procedure SetMenuStates;

  public
    constructor Create;
    destructor Destroy; override; 
    function MsgProc(hWnd: HWND; uMsg: LongWord; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;



implementation

uses SysUtils, Classes, Math;


{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor.   Paired with ~CMyD3DApplication()
//       Member variables should be initialized to a known state here.  
//       The application window has not yet been created and no Direct3D device 
//       has been created, so any initialization that depends on a window or 
//       Direct3D should be deferred to a later stage. 
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;
  m_Camera:= CFirstPersonCamera.Create;
  m_DSound:= CSoundManager.Create;
  m_MultiAnim:= CMultiAnim.Create;
  m_v_pCharacters:= TTinyList.Create;

  // inherited
  m_strWindowTitle := 'MultiAnimation';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  // intrinsic
  m_pMeshFloor := nil;
  m_pTxFloor := nil;

  m_pFont := nil;

  m_dwFollow := $ffffffff;
  m_bDrawReflection := True;
  m_bDisplayText := True;
  m_bDisplayHelp := True;

  m_fLastAnimTime := 0.0;

  m_dwCreationWidth := 640;
  m_dwCreationHeight := 480;

  m_bPlaySounds := True;
end;


destructor CMyD3DApplication.Destroy;
begin
  FreeAndNil(m_Camera);
  FreeAndNil(m_DSound);
  FreeAndNil(m_MultiAnim);
  FreeAndNil(m_v_pCharacters);
  inherited;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;
  // Debugging vertex shaders requires either REF or software vertex processing
  // and debugging pixel shaders requires REF.
  {$IFDEF DEBUG_VS}
  if (pCaps.DeviceType <> D3DDEVTYPE_REF) and
     ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then Exit;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
  if (pCaps.DeviceType <> D3DDEVTYPE_REF) then Exit;
  {$ENDIF}

  // If device doesn't support 1.1 vertex shaders in HW, switch to SWVP.
  if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,1)) then
  begin
    if ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then Exit;
  end;

  // Need to support A8R8G8B8 textures
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
                                     adapterFormat, 0,
                                     D3DRTYPE_TEXTURE, D3DFMT_A8R8G8B8))
  then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Paired with FinalCleanup().
//       The window has been created and the IDirect3D9 interface has been
//       created, but the device has not been created yet.  Here you can
//       perform application-related initialization and cleanup that does
//       not depend on a device.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HRESULT;
var
  vEye: TD3DXVector3;
  vAt: TD3DXVector3;
begin
  Randomize;

  // Setup the camera with view matrix
  vEye:= D3DXVector3(0.5, 0.55, -0.2);
  vAt:= D3DXVector3(0.5,  0.125, 0.5);
  m_Camera.SetViewParams(vEye, vAt);
  m_Camera.SetScalers(0.01, 1.0);  // Camera movement parameters

  // set up dsound
  m_DSound.Initialize(m_hWnd, DSSCL_PRIORITY);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Paired with DeleteDeviceObjects()
//       The device has been created.  Resources that are not lost on
//       Reset() can be created here -- resources in D3DPOOL_MANAGED,
//       D3DPOOL_SCRATCH, or D3DPOOL_SYSTEMMEM.  Image surfaces created via
//       CreateImageSurface are never lost and can be created here.  Vertex
//       shaders and pixel shaders can also be created here as they are not
//       lost on Reset().
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HRESULT;
var
  hDC: Windows.HDC;
  nHeight: Integer;
  tszPath, sXFile, sFXFile: array[0..MAX_PATH-1] of Char;
  mx: TD3DXMatrix;
  AH: CMultiAnimAllocateHierarchy;
  caps: TD3DCaps9;
  i: Integer;
  pTiny: CTiny;
begin
  // font stuff
  hDC := GetDC(0);
  nHeight := -MulDiv(10, GetDeviceCaps(hDC, LOGPIXELSY), 72);
  ReleaseDC(0, hDC);
  Result := D3DXCreateFont(m_pd3dDevice, nHeight, 0, FW_BOLD, 0, False, DEFAULT_CHARSET,
                           OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH or FF_DONTCARE,
                           'Arial', m_pFont);
  if FAILED(Result) then Exit;


  // Initialize floor textures
  Result := DXUtil_FindMediaFileCch(tszPath, MAX_PATH, TXFILE_FLOOR);
  if FAILED(Result) then Exit;

  Result := D3DXCreateTextureFromFile(m_pd3dDevice, tszPath, m_pTxFloor);
  if FAILED(Result) then Exit;

  // floor geometry transform
  D3DXMatrixRotationX(m_mxFloor, -D3DX_PI / 2.0);
  D3DXMatrixRotationY(mx, D3DX_PI / 4.0);
  D3DXMatrixMultiply(m_mxFloor, m_mxFloor, mx);
  D3DXMatrixTranslation(mx, 0.5, 0.0, 0.5);
  D3DXMatrixMultiply(m_mxFloor, m_mxFloor, mx);

  // set material for floor
  m_MatFloor.Diffuse := D3DXColor(1.0, 1.0, 1.0, 0.75);
  m_MatFloor.Ambient := D3DXColor(1.0, 1.0, 1.0, 1.0);
  m_MatFloor.Specular := D3DXColor(0.0, 0.0, 0.0, 1.0);
  m_MatFloor.Emissive := D3DXColor(0.0, 0.0, 0.0, 0.0);
  m_MatFloor.Power := 0.0;

  SetMenuStates;

  // set up MultiAnim
  Result := DXUtil_FindMediaFileCch(sXFile, MAX_PATH, 'tiny_4anim.x');
  if FAILED(Result) then Exit;
  Result := DXUtil_FindMediaFileCch( sFXFile, MAX_PATH, 'MultiAnimation.fx');
  if FAILED(Result) then Exit;

  AH:= CMultiAnimAllocateHierarchy.Create;
  try
    AH.SetMA(m_MultiAnim);

    Result := m_MultiAnim.Setup(m_pd3dDevice, sXFile, sFXFile, AH);
    if FAILED(Result) then Exit;

    // get device caps
    m_pd3dDevice.GetDeviceCaps(caps);

    // Select the technique that fits the shader version.
    // We could have used ValidateTechnique()/GetNextValidTechniqe() to find the
    // best one, but this is convenient for our purposes.
    if (caps.VertexShaderVersion <= D3DVS_VERSION(1,1)) then
    begin
      if (caps.VertexShaderVersion < D3DVS_VERSION(1,1)) then
      begin
        m_MultiAnim.SetTechnique('t2');
        m_MultiAnim.SetSWVP(True);
      end else
      begin
        m_MultiAnim.SetTechnique('t1');
        m_MultiAnim.SetSWVP(False);
      end;
    end else
    begin
      m_MultiAnim.SetTechnique('t0');
      m_MultiAnim.SetSWVP(False);
    end;

    // Restore steps for tiny instances
    for i:= 0 to m_v_pCharacters.Count - 1 do
    begin
      m_v_pCharacters[i].RestoreDeviceObjects(m_pd3dDevice);
    end;

    // If there is no instance, make sure we have at least one.
    if (m_v_pCharacters.Count = 0) then
    begin
      try
        pTiny := CTiny.Create;
      except
        Result:= E_OUTOFMEMORY;
        Exit;
      end;

      Result := pTiny.Setup(m_MultiAnim, m_v_pCharacters, m_DSound, 0.0);
    end;
  finally
    AH.Free;
  end;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Paired with InvalidateDeviceObjects()
//       The device exists, but may have just been Reset().  Resources in
//       D3DPOOL_DEFAULT and any other device state that persists during
//       rendering should be set here.  Render states, matrices, textures,
//       etc., that don't change during rendering can be set once here to
//       avoid redundant state setting during Render() or FrameMove().
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  pMesh: ID3DXMesh;
  dwNumVx: DWORD;
type
  PVx = ^TVx;
  TVx = packed record
    vPos: TD3DXVector3;
    vNorm: TD3DXVector3;
    fTex: array[0..1] of Single;
  end;
var
  Vx: PVx;
  i: Integer;
begin
  // set lighting
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT, D3DCOLOR_ARGB( 255, 255, 255, 255 ));
  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_NORMALIZENORMALS, iTrue);

  // Set up the camera's projection matrix
  m_fAspectRatio := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  m_Camera.SetProjParams(D3DXToRadian(60.0), m_fAspectRatio, 0.001, 100.0);

  // create the floor geometry
  Result := D3DXCreatePolygon(m_pd3dDevice, 1.2, 4, pMesh, nil);
  if FAILED(Result) then Exit;

  Result := pMesh.CloneMeshFVF(D3DXMESH_WRITEONLY, D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1,
                               m_pd3dDevice, m_pMeshFloor);
  pMesh:= nil;
  if FAILED(Result) then Exit;

  dwNumVx := m_pMeshFloor.GetNumVertices;

  // Initialize its texture coordinates
  Result := m_pMeshFloor.LockVertexBuffer(0, Pointer(Vx));
  if FAILED(Result) then Exit;

  for i := 0 to dwNumVx - 1 do
  begin
    if (Abs(Vx.vPos.x) < 0.01) then
    begin
      if (Vx.vPos.y > 0) then
      begin
        Vx.fTex[ 0 ] := 0.0;
        Vx.fTex[ 1 ] := 0.0;
      end else
      if (Vx.vPos.y < 0.0) then
      begin
        Vx.fTex[ 0 ] := 1.0 * FLOOR_TILECOUNT;
        Vx.fTex[ 1 ] := 1.0 * FLOOR_TILECOUNT;
      end else
      begin
        Vx.fTex[ 0 ] := 0.5 * FLOOR_TILECOUNT;
        Vx.fTex[ 1 ] := 0.5 * FLOOR_TILECOUNT;
      end;
    end else
    if (Vx.vPos.x > 0.0) then
    begin
      Vx.fTex[ 0 ] := 1.0 * FLOOR_TILECOUNT;
      Vx.fTex[ 1 ] := 0.0;
    end else
    begin
      Vx.fTex[ 0 ] := 0.0;
      Vx.fTex[ 1 ] := 1.0 * FLOOR_TILECOUNT;
    end;

    Inc(Vx);
  end;

  m_pMeshFloor.UnlockVertexBuffer;

  // Restore font
  if Assigned(m_pFont) then m_pFont.OnResetDevice;

  // reset the timer
  m_fLastAnimTime := m_fTime;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.  Paired with RestoreDeviceObjects()
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  if Assigned(m_pFont) then m_pFont.OnLostDevice;

  SAFE_RELEASE(m_pMeshFloor);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Paired with InitDeviceObjects()
//       Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
var
  i: Integer;
  AH: CMultiAnimAllocateHierarchy;
begin
  SAFE_RELEASE(m_pFont);

  for i:= 0 to m_v_pCharacters.Count - 1 do
    m_v_pCharacters[i].InvalidateDeviceObjects;

  AH:= CMultiAnimAllocateHierarchy.Create;
  try
    AH.SetMA(m_MultiAnim);
    m_MultiAnim.Cleanup(AH);
  finally
    AH.Free;
  end;

  SAFE_RELEASE(m_pTxFloor);
  SAFE_RELEASE(m_pMeshFloor);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Paired with OneTimeSceneInit()
//       Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
var
  i: Integer;
begin
  for i:= 0 to m_v_pCharacters.Count - 1 do
    m_v_pCharacters[i].Cleanup;
  m_v_pCharacters.Clear;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
var
  i: Integer;
begin
  // we only need to call the animation routine 60 times per second
  if (m_fTime - m_fLastAnimTime > (1.0 / 60.0)) then
  begin
    for i:= 0 to m_v_pCharacters.Count - 1 do
      m_v_pCharacters[i].Animate(m_fTime - m_fLastAnimTime);

    m_fLastAnimTime := m_fTime;
  end;

  // Update the camera
  m_Camera.FrameMove(m_fElapsedTime);

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
var
  mx, mxView, mxProj: TD3DXMatrix;
  vEye: TD3DXVector3;
  vLightDir: TD3DXVector3;
  pChar: CTiny;
  vCharPos: TD3DXVector3;
  vCharFacing: TD3DXVector3;
  vAt, vUp: TD3DXVector3;
  pEffect: ID3DXEffect;
  v: TD3DXVector4;
  i: Integer;
  rc: TRect;
  sDesc: array[0..255] of Char;
  v_sReport: TStringList;
begin
{$IFDEF UNICODE}
  For UNICODE you should use "m_pFont.DrawTextW" routine!!!
{$ENDIF}

  // clear
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $003fafff, 1.0, 0);

  // Render the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // set up the camera

    // are we following a tiny, or an independent arcball camera?
    if (m_dwFollow = $ffffffff) then
    begin
      // Light direction is same as camera front (reversed)
      // vLightDir :=  -m_Camera.GetWorldAhead;
      D3DXVec3Scale(vLightDir, m_Camera.GetWorldAhead, -1);

      // set static transforms
      mxView := m_Camera.GetViewMatrix^;
      mxProj := m_Camera.GetProjMatrix^;
      m_pd3dDevice.SetTransform(D3DTS_VIEW, mxView);
      m_pd3dDevice.SetTransform(D3DTS_PROJECTION, mxProj);
      vEye := m_Camera.GetEyePt;
    end else
    begin
      // set follow transforms
      pChar := m_v_pCharacters[m_dwFollow];

      pChar.GetPosition(vCharPos);
      pChar.GetFacing(vCharFacing);
      vEye := D3DXVector3(vCharPos.x, 0.25, vCharPos.z);
      vAt:= D3DXVector3(vCharPos.x, 0.0125, vCharPos.z);
      vUp:= D3DXVector3(0.0, 1.0, 0.0);
      vCharFacing.x := vCharFacing.x * 0.25; vCharFacing.y := 0.0; vCharFacing.z := vCharFacing.z * 0.25;
      // vEye := vEye - vCharFacing;
      D3DXVec3Subtract(vEye, vEye, vCharFacing);
      // vAt  := vAt  + vCharFacing;
      D3DXVec3Add(vAt, vAt, vCharFacing);

      D3DXMatrixLookAtLH(mxView, vEye, vAt, vUp);
      m_pd3dDevice.SetTransform(D3DTS_VIEW, mxView);

      D3DXMatrixPerspectiveFovLH(mxProj, D3DXToRadian(60.0), m_fAspectRatio, 0.1, 100.0);
      m_pd3dDevice.SetTransform(D3DTS_PROJECTION, mxProj);

      // Set the light direction and normalize
      D3DXVec3Subtract(vLightDir, vEye, vAt);
      D3DXVec3Normalize(vLightDir, vLightDir);
    end;

    pEffect := m_MultiAnim.GetEffect;
    if (pEffect <> nil) then
    begin
      // set view-proj
      D3DXMatrixMultiply(mx, mxView, mxProj);
      pEffect.SetMatrix('mViewProj', mx);

      // Set the light direction so that the
      // visible side is lit.
      v:= D3DXVector4(vLightDir.x, vLightDir.y, vLightDir.z, 1.0);
      pEffect.SetVector('lhtDir', v);

      pEffect:= nil;
    end;

    // set the fixed function shader for drawing the floor
    m_pd3dDevice.SetVertexShader(nil);
    m_pd3dDevice.SetFVF(m_pMeshFloor.GetFVF);

    // draw the floor
    m_pd3dDevice.SetTexture(0, m_pTxFloor);
    m_pd3dDevice.SetMaterial(m_MatFloor);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, m_mxFloor);
    m_pMeshFloor.DrawSubset(0);

    // draw each tiny
    for i:= 0 to m_v_pCharacters.Count - 1 do
    with m_v_pCharacters[i] do
    begin
      // set the time to update the hierarchy
      AdvanceTime(m_fElapsedTime, @vEye);
      // draw the mesh
      Draw;
    end;

    //
    // Output text information
    //
    if Assigned(m_pFont) then
    begin
      // Dump out the FPS and device stats

      SetRect(rc, 2, 0, 0, 0);
      m_pFont.DrawTextA(nil, m_strFrameStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));
      Inc(rc.Top, 16);
      m_pFont.DrawTextA(nil, m_strDeviceStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));
      Inc(rc.Top, 16);
      StrFmt(sDesc, '  Time: %2.3f', [m_fTime]);
      m_pFont.DrawTextA(nil, sDesc, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
      Inc(rc.Top, 16);
      StrFmt(sDesc, '  Number of models: %d', [m_v_pCharacters.Count]);
      m_pFont.DrawTextA(nil, sDesc, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));

      // We can only display either the behavior text or help text,
      // with the help text having priority.
      if m_bDisplayHelp then
      begin
        Inc(rc.Top, 16);
        if (m_dwFollow = $FFFFFFFF) then
          m_pFont.DrawTextA(nil, 'Press F1 to hide help', -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)))
        else
          // Following an instance. Let user know F1 shows animation info
          m_pFont.DrawTextA(nil, 'Press F1 to show animation info', -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));

        rc.left := 20;
        rc.top := m_d3dsdBackBuffer.Height - 165;
        m_pFont.DrawTextA(nil,
                          'Global Controls:'#10 +
                          'Add instance'#10 +
                          'Next view'#10 +
                          'Previous view'#10 +
                          'Reset camera'#10 +
                          'Reset time'#10 +
                          'Toggle behavior text'#10 +
                          'Change device'#10 +
                          'Quit'#10 +
                          'Relinquish control for all instances',
                          -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
        rc.left := 250;
        m_pFont.DrawTextA(nil,
                          ''#10 +
                          'Z'#10 +
                          'N'#10 +
                          'P'#10 +
                          'R'#10 +
                          'I'#10 +
                          'T'#10 +
                          'F2'#10 +
                          'Esc'#10 +
                          'V',
                          -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
        rc.top := m_d3dsdBackBuffer.Height - 85;
        rc.left := 370;
        m_pFont.DrawTextA(nil,
                          'For a specific instance:'#10 +
                          'Take control'#10 +
                          'Move forward'#10 +
                          'Turn'#10 +
                          'Enable run mode'#10,
                          -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
        rc.left := 520;
        m_pFont.DrawTextA(nil,
                          ''#10 +
                          'C'#10 +
                          'W'#10 +
                          'A,D'#10 +
                          'Shift'#10,
                          -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));

      end else
      if m_bDisplayText then
      begin
        Inc(rc.Top, 16);
        m_pFont.DrawTextA(nil, 'Press F1 to hide help', -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));

        // output data for T[m_dwFollow]
        if (m_dwFollow <> $ffffffff) then
        begin
          pChar := m_v_pCharacters[m_dwFollow];
          v_sReport:= TStringList.Create;
          try
            pChar.Report(v_sReport);
            rc.left := 5;
            rc.top := m_d3dsdBackBuffer.Height - 115;
            for i := 0 to 5 do
            begin
              m_pFont.DrawTextA(nil, PAnsiChar(v_sReport[i]), -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
              Inc(rc.Top, 16);
            end;
            m_pFont.DrawTextA(nil, PAnsiChar(v_sReport[16]), -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));

            rc.left := 220;
            rc.top := m_d3dsdBackBuffer.Height - 85;
            for i := 6 to 10 do
            begin
              m_pFont.DrawTextA(nil, PAnsiChar(v_sReport[i]), -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
              Inc(rc.Top, 16);
            end;

            rc.left := 390;
            rc.top := m_d3dsdBackBuffer.Height - 85;
            for i := 11 to 15 do
            begin
              m_pFont.DrawTextA(nil, PAnsiChar(v_sReport[i]), -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
              Inc(rc.Top, 16);
            end;
          finally
            v_sReport.Free;
          end;
        end;
      end;
    end;

    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: LongWord;
  wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  i: Integer;
  hr: HRESULT;
  pTiny: CTiny;  
begin
  // Pass messages to camera class for camera movement if the
  // global camera if active
  if (DWORD(-1) = m_dwFollow) then m_Camera.HandleMessages(hWnd, uMsg, wParam, lParam);

  // Handle key presses
  case uMsg of
    WM_COMMAND:
    begin
      case LOWORD(wParam) of
        IDM_TOGGLEHELP:        m_bDisplayHelp := not m_bDisplayHelp;
        ID_OPTIONS_REFLECTION: m_bDrawReflection := not m_bDrawReflection;

        ID_OPTIONS_TAKECONTROL:
          if (m_dwFollow <> $ffffffff) then
          begin
            m_v_pCharacters[m_dwFollow].SetUserControl;
          end;

        ID_OPTIONS_RESETTIME:
        begin
          DXUtil_Timer(TIMER_RESET);
          m_fTime := DXUtil_Timer(TIMER_GETAPPTIME);
          m_fLastAnimTime := m_fTime;
          for i:= 0 to m_v_pCharacters.Count - 1 do
            m_v_pCharacters[i].ResetTime;
        end;

        ID_OPTIONS_TOGGLESOUNDS:
        begin
          m_bPlaySounds := not m_bPlaySounds;
          for i:= 0 to m_v_pCharacters.Count - 1 do
            m_v_pCharacters[i].SetSounds(m_bPlaySounds);
        end;

        IDM_ADDTINY:
        begin
          pTiny := CTiny.Create;
          hr := pTiny.Setup(m_MultiAnim, m_v_pCharacters, m_DSound, m_fTime);
          if SUCCEEDED(hr)
            then pTiny.SetSounds(m_bPlaySounds)
            else pTiny.Free;
        end;

        ID_VIEW_TOGGLEBEHAVIORTEXT: m_bDisplayText := not m_bDisplayText;

        ID_VIEW_RESETCAMERA: m_dwFollow := $ffffffff;

        ID_VIEW_PREVIOUSVIEW:
          if (m_v_pCharacters.Count <> 0) then
          begin
            if (m_dwFollow = $ffffffff) then
              m_dwFollow := m_v_pCharacters.Count - 1
            else if (m_dwFollow = 0) then
              m_dwFollow := $ffffffff
            else
              Dec(m_dwFollow);
          end;

        ID_VIEW_NEXTVIEW:
          if (m_v_pCharacters.Count <> 0) then
          begin
            if (m_dwFollow = $ffffffff) then
              m_dwFollow := 0
            else if (Integer(m_dwFollow) = m_v_pCharacters.Count - 1) then
              m_dwFollow := $ffffffff
            else
              Dec(m_dwFollow);
          end;
      end;

      // Update the menus, in case any state changes occurred
      SetMenuStates;
    end;
  end;

  // Pass remaining messages to default handler
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;


//-----------------------------------------------------------------------------
// Name: SetMenuStates()
// Desc:
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.SetMenuStates;
var
  hMenu: Windows.HMENU;
begin
  hMenu := GetMenu(m_hWnd);

  CheckMenuItem(hMenu, ID_OPTIONS_REFLECTION, IfThen(m_bDrawReflection, MF_CHECKED, MF_UNCHECKED));
  EnableMenuItem(hMenu, ID_OPTIONS_TAKECONTROL, IfThen(m_dwFollow <> $ffffffff, MF_ENABLED, MF_GRAYED));
end;

end.

